/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2019 IndiScale GmbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */

'use strict';

var autocomplete = new function() {
	this.version = "0.1";

	/**
	 * @type {Promise} promise for a list of entity names.
	 */
	this.candidates = undefined;
	var logger = log.getLogger("autocomplete");
	this.init = function() {
		this.candidates = undefined;
		this.toggle_completion();
		logger.setLevel("trace");
	};

	this.retrieve_names = async function () {
		var response = $(await connection.get(transaction.generateEntitiesUri(
			["names"]))).find("Response [name]")

		return response.toArray().map(x=> $(x).attr("name"));
	};

	/**
	 * @return {Promise} promise for a list of entity names.
	 */
	this.get_candidates = function () {

		if (typeof this.candidates === "undefined"){

			this.candidates = this.retrieve_names();
		};
		return this.candidates;
	};

	this.filter = function (x, qry){
		return x.toLowerCase().startsWith(qry.toLowerCase())
	};

	this.search = async function (qry, callback, origJQElement) { 
		const names = await autocomplete.get_candidates();
		callback(names.filter(x => autocomplete.filter(x, qry)));
	};

	this.typed = function (newValue, origJQElement) { 
		var cursorpos = origJQElement[0].selectionEnd;
		var beginning_of_word = newValue.slice(0, cursorpos).lastIndexOf(" ");
		var word = newValue.slice(beginning_of_word+1, cursorpos);
		return  word;
	};

	this.searchPost = function (resultsFromServer, origJQElement) { 
		var cursorpos = origJQElement[0].selectionEnd;
		var newValue = origJQElement[0].value
		var beginning_of_word = newValue.slice(0, cursorpos).lastIndexOf(
			" ");
		var start = newValue.slice(0, beginning_of_word+1);
		var end = origJQElement[0].value.slice(cursorpos);
		var result = resultsFromServer.map( x => {
			return {
				text: start + x + end, 
				html: x
			}});
		return result;
	};

    this.toggle_completion = function () {
		var field  = $("#caosdb-query-textarea");
		field.toggleClass("basicAutoComplete", true);
		field.autoComplete({
				events: {
					search: this.search,
		 			typed: this.typed,
		 			searchPost: this.searchPost,
				}
			}
		);
		return field;
	};
};

//$(document).ready(function () {
//   caosdb_modules.register(autocomplete);
//});
